from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

import cv2
import numpy as np
import scipy.ndimage as ndi
import tifffile
from PIL import Image

from ..base import MaskDataset

if TYPE_CHECKING:
    import albumentations


class BBBC039(MaskDataset):
    """Nuclei of U2OS cells in a chemical screen [1]_

    This data set has a total of 200 fields of view of nuclei captured with
    fluorescence microscopy using the Hoechst stain. These images are a sample
    of the larger BBBC022 chemical screen. The images are stored as TIFF files
    with 520x696 pixels at 16 bits.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    training : bool, default: True
        Load training set if True, else load testing one

    Notes
    -----
    - Split (training/valiadation/test)
        - `training=True` combines 'training' with 'validation'
    - Annotate objs not touching each other with 1 and use 2, 3, ... for the
      touching ones. It is great and clever, but it does not follow the form of
      other instance segmented masks. ``get_mask()`` will make a instance
      labeled mask (each obj has unique labels). After labeling max label is 231
      for training, and 202 for test. So having masks of dtype UINT8 is fine.
    - Max label is 3 (in original annotation)
    - Sample of larger BBBC022 and did manual segmentation
    - Possible overlap some with DSB2018
    - Mask is png but (instance) value is only stored in RED channel
    - Maximum value is 2**12

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC039

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """

    # Dataset's acronym
    acronym = 'BBBC039'
    _max_val = 4095  # 2**12

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        # specific to this dataset
        training: bool = True,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_samples = num_samples
        # specific to this dataset
        self.training = training

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        img = img / np.float32(self._max_val)
        return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)

    def get_mask(self, p: Path) -> np.ndarray:
        mask = np.asarray(Image.open(p))[..., 0]
        max_val = mask.max()
        if max_val == 0:
            return mask
        else:
            inst_mask, n_labels = ndi.label(mask == 1, output='uint8')  # int32 by default
            for m in range(2, max_val+1):
                labeled, n = ndi.label(mask == m, output='uint8')
                inst_mask += np.where(labeled == 0, 0, labeled + n_labels)
                n_labels += n
        return inst_mask

    @cached_property
    def file_list(self) -> List[Path]:
        root_dir = self.root_dir
        parent = root_dir / 'images'
        file_list = []
        for name in self.ids:
            p = parent / name
            file_list.append(p.with_suffix('.tif'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        parent = root_dir / 'masks'
        anno_list = []
        for name in self.ids:
            p = parent / name
            anno_list.append(p)
        return dict((k, v) for k, v in enumerate(anno_list))

    @cached_property
    def ids(self) -> list:
        def _readlines(path):
            with open(path, 'r') as f:
                lines = f.readlines()
            return list(map(lambda s: s.strip(), lines))
        meta_dir = self.root_dir / 'metadata'
        if self.training:
            # Combine training and validation
            meta_file = meta_dir / 'training.txt'
            _ids = _readlines(meta_file)
            meta_file = meta_dir / 'validation.txt'
            _ids += _readlines(meta_file)
        else:
            meta_file = meta_dir / 'test.txt'
            _ids = _readlines(meta_file)
        return _ids



"""Define a base class and its interface

``Dataset`` is the base of all datasets

``MaskDataset`` is the base of datasets that have mask annotation
"""

import abc
import inspect
import random
from pathlib import Path
from typing import (TYPE_CHECKING, Any, Dict, Iterator, Optional, Sequence,
                    Union)

import cv2
import numpy as np

if TYPE_CHECKING:
    import albumentations

class DatasetInterface(metaclass=abc.ABCMeta):
    """Dataset interface

    Attributes
    ----------
    __repr__
    __len__
    __getitem__
    acronym
    root_dir
    file_list

    Methods
    -------
    get_image

    """

    @abc.abstractmethod
    def __repr__(self):  # common
        """Print info of dataset"""
        ...

    @abc.abstractmethod
    def __len__(self):  # common
        ...

    @abc.abstractmethod
    def __getitem__(self, ind):  # common
        """Given index returns a dictionary of key(s) and array"""
        ...

    @abc.abstractmethod
    def get_image(self, key):  # required
        """Get an image"""
        ...

    @classmethod
    @property
    @abc.abstractmethod
    def acronym(cls):  # required
        """Assign acroym for a subclass dataset"""
        ...

    @property
    @abc.abstractmethod
    def root_dir(self):  # required
        """Path to root directory"""
        ...

    @property
    def file_list(self):  # required
        """A list of pathes to image files"""
        ...


class Dataset(DatasetInterface):
    """Base to define common attributes and methods for [`MaskDataset`, ...]

    Attributes
    ----------
    __repr__
    __len__
    __iter__
    root_dir
    output
    transforms
    num_samples
    grayscale : optional
    grayscale_mode : optional
    num_channels : optional

    Methods
    -------
    __getitem__
    _drop_missing_pairs
    to_gray

    Notes
    -----
    Required attributes in subclass
        - ``anno_dict``
        - ``__getitem__()``
        - ``get_image()``

    """

    def __repr__(self):
        signature = inspect.signature(self.__init__)
        params = list(signature.parameters.keys())
        # remove 'kwargs'. try/except would be a better choice.
        if 'kwargs' in params:
            params.remove('kwargs')
        init_args_str = '(' + ', '.join(f'{k}={getattr(self, k)}' for k in params) + ')'
        return self.acronym + init_args_str

    def __len__(self):
        """Length of dataset. Can be overwritten with ``num_samples``"""
        if self.num_samples is not None:
            return self.num_samples
        return len(self.file_list)

    def __iter__(self):
        return IterDataset(self)

    @property
    def root_dir(self) -> Path:
        if hasattr(self, '_root_dir'):
            _root_dir = getattr(self, '_root_dir')
            if not isinstance(_root_dir, Path):
                return Path(_root_dir)
            return _root_dir
        raise NotImplementedError("Attr `_root_dir` not defined")

    @property
    def output(self) -> str:
        """Determine return(s) when called, fixed to 'image'"""
        return 'image'

    @property
    def transforms(self) -> Optional['albumentations.Compose']:
        """Transform images and masks"""
        if hasattr(self, '_transforms'):
            return getattr(self, '_transforms')
        return None

    @property
    def num_samples(self) -> Optional[int]:
        """Number of calls that will override __len__"""
        if hasattr(self, '_num_samples'):
            return getattr(self, '_num_samples')
        return None

    @num_samples.setter
    def num_samples(self, val):
        self._num_samples = val

    @property
    def grayscale(self) -> Optional[bool]:
        """Flag for grayscale conversion"""
        if hasattr(self, '_grayscale'):
            return getattr(self, '_grayscale')
        return None

    @grayscale.setter
    def grayscale(self, val):
        self._grayscale = val

    @property
    def grayscale_mode(self) -> Optional[Union[str, Sequence[float]]]:
        """Determine grayscale mode one of {'cv2', 'equal', Sequence[float]}
        """
        if hasattr(self, '_grayscale_mode'):
            return getattr(self, '_grayscale_mode')
        return None

    @grayscale_mode.setter
    def grayscale_mode(self, val):
        self._grayscale_mode = val

    @property
    def num_channels(self) -> Optional[int]:
        """Number of image channels used for `to_gray()`"""
        if hasattr(self, '_num_channels'):
            return getattr(self, '_num_channels')
        return None

    def __getitem__(self, ind: int) -> Dict[str, np.ndarray]:
        """Get image

        Dataset does not any annotation available. It will only load 'image'.

        Parameters
        ----------
        ind : int
            Index to get path(s) from ``file_list`` attribute

        Attributes
        ----------
        self.file_list

        Other Parameters
        ----------------
        self._transforms
        self._num_samples
        self._grayscale
        self._grayscale_mode
        self._num_channels

        """
        # Randomize `ind` when `num_samples` set
        if self.num_samples is not None:
            if ind >= self.num_samples:
                raise IndexError('list index out of range')
            ind_max = len(self.file_list)
            ind = random.randrange(0, ind_max)
        # `output="image"`
        p = self.file_list[ind]
        image = self.get_image(p)
        if self.grayscale:
            num_channels = self.num_channels
            # exception
            if hasattr(self, 'image_ch') and len(self.image_ch) == 1:
                # BBBC020: `_num_channels=2`. When `image_ch` is set to one
                # channel, output images become grayscale and `to_gray()`
                # got `num_channels=2`.
                num_channels = 3
            if num_channels is None:
                num_channels = len(p) if isinstance(p, list) else 3
            image = self.to_gray(
                image,
                grayscale_mode=self.grayscale_mode,
                num_channels=num_channels
            )
        if self.transforms:
            image = self.transforms(image=image)['image']
        return {'image': image}

    def _drop_missing_pairs(self) -> tuple:
        """Drop images and reindex the anno list (dict)

        Sometimes, not all images have annotation. For consistence, this func
        simply drops those images missing annotation.

        For example,
        - MurphyLab
        - BBBC018
        - BBBC020
        """
        file_list = getattr(self, 'file_list')
        anno_dict = getattr(self, 'anno_dict')
        _diff = set(range(len(file_list))).difference(set(anno_dict))
        diff = sorted(_diff)
        # logger.info(f'{self.acronym}:Dropping indices: {diff}')
        for i, ind in enumerate(diff):
            file_list.pop(ind-i)
        anno_dict = dict((i, v) for i, v in enumerate(anno_dict.values()))
        return file_list, anno_dict

    @staticmethod
    def to_gray(
        arr: np.ndarray,
        grayscale_mode: Optional[Union[str, Sequence[float]]] = None,
        num_channels: int = 3,
    ) -> np.ndarray:
        """Convert bioimage to grayscale

        Parameters
        ----------
        arr : image array
            Numpy image array whose shape is (h, w, 3)
        grayscale_mode : str or sequence of float, optional
            Choose a strategy for gray conversion. Three options are availble.
            Either one of {'cv2', 'equal'} or be a sequence of float numbers,
            which indicate linear weights of each channel.
        num_channels : int
            Explicitly set number of channels for `grayscale_mode='equal'`.
        """
        if isinstance(grayscale_mode, str):
            if grayscale_mode == 'cv2':
                if arr.shape[-1] != 3:
                    raise ValueError("Image arr should have RGB channels")
                arr = cv2.cvtColor(arr, cv2.COLOR_RGB2GRAY)
                arr = cv2.cvtColor(arr, cv2.COLOR_GRAY2RGB)
            elif grayscale_mode == 'equal':
                # Expect (h, w, ch) shape of array
                arr = (arr.sum(axis=-1) / num_channels).astype(arr.dtype)
                arr = cv2.cvtColor(arr, cv2.COLOR_GRAY2RGB)
            else:
                raise ValueError(f"Wrong `grayscale_mode={grayscale_mode}`")
        else:
            raise NotImplementedError("`grayscale_mode`")
        return arr


class MaskDataset(Dataset):
    """Base for datasets with mask annotation

    Define ``__getitem__`` method to load mask annotation paired with image.
    Pre-defined attributes are prefixed with a single underscore to distinguish
    them from those specific to a dataset. It is required to implement two
    methods: ``get_image()`` and ``get_mask()`` as well as ``acronym`` and
    ``_root_dir`` for each subclass.

    Attributes
    ----------
    output
    anno_dict

    Methods
    -------
    __getitem__

    Notes
    -----
    Required attributes in subclass
        - ``acronym``
        - ``_root_dir``
        - ``_output``
        - ``_grayscale`` (optional)
        - ``_grayscale_mode`` (optional)
        - ``_num_channels`` (optional)
        - ``get_image()``
        - ``get_mask()`` (optional)

    See Also
    --------
    Dataset : super class

    """
    @property
    def output(self) -> str:
        """Determine return(s) when called"""
        return self._output

    @output.setter
    def output(self, val):
        self._output = val

    @property
    def anno_dict(self) -> Dict[int, Any]:
        """Dictionary of pathes to annotation files"""
        raise NotImplementedError

    def __getitem__(self, ind: int) -> Dict[str, np.ndarray]:
        """Get image, mask, or both depending on ``output`` argument

        For MaskDataset, available output types are ['image', 'mask', 'both'].

        Parameters
        ----------
        ind : int
            Index to get path(s) from ``file_list`` attribute

        Attributes
        ----------
        self.output
        self.file_list
        self.anno_dict

        Other Parameters
        ----------------
        self._transforms
        self._num_samples
        self._grayscale
        self._grayscale_mode
        self._num_channels

        """
        # Randomize `ind` when `num_samples` set
        if self.num_samples is not None:
            if ind >= self.num_samples:
                raise IndexError('list index out of range')
            ind_max = len(self.file_list)
            if (self.output != 'image') and (self.anno_dict is not None):
                ind_max = len(self.anno_dict)
            ind = random.randrange(0, ind_max)
        # `output="image"`
        if self.output == 'image':
            p = self.file_list[ind]
            image = self.get_image(p)
            if self.grayscale:
                num_channels = self.num_channels
                # exception
                if hasattr(self, 'image_ch') and len(self.image_ch) == 1:
                    # BBBC020: `_num_channels=2`. When `image_ch` is set to one
                    # channel, output images become grayscale and `to_gray()`
                    # got `num_channels=2`.
                    num_channels = 3
                if num_channels is None:
                    num_channels = len(p) if isinstance(p, list) else 3
                image = self.to_gray(
                    image,
                    grayscale_mode=self.grayscale_mode,
                    num_channels=num_channels
                )
            if self.transforms:
                image = self.transforms(image=image)['image']
            return {'image': image}
        # `output="mask"`
        elif self.output == 'mask':
            pm = self.anno_dict[ind]
            mask = self.get_mask(pm)
            _image = np.zeros_like(mask, dtype=np.uint8)  # dummy image
            if self.transforms is not None:
                mask = self.transforms(image=_image, mask=mask)['mask']
                # # Filtering out empty masks
                # while mask.max() == 0:
                #     mask = self.transforms.augment_image(mask)
            return {'mask': mask}
        # both image and gt
        elif self.output == 'both':
            # 'image'
            p = self.file_list[ind]
            image = self.get_image(p)
            # 'mask'
            pm = self.anno_dict[ind]
            mask = self.get_mask(pm)
            if self.grayscale:
                num_channels = self.num_channels
                # exception
                if hasattr(self, 'image_ch') and len(self.image_ch) == 1:
                    # BBBC020: `_num_channels=2`. When `image_ch` is set to one
                    # channel, output images become grayscale and `to_gray()`
                    # got `num_channels=2`.
                    num_channels = 3
                if num_channels is None:
                    num_channels = len(p) if isinstance(p, list) else 3
                image = self.to_gray(
                    image,
                    grayscale_mode=self.grayscale_mode,
                    num_channels=num_channels
                )
            # Make sure to apply the same augmentation both to image and mask
            if self.transforms is not None:
                augmented = self.transforms(image=image, mask=mask)
                image, mask = augmented['image'], augmented['mask']
            return {'image': image, 'mask': mask}
        else:
            raise NotImplementedError("Choose one ['image', 'mask', 'both']")

    def get_mask(self, key) -> np.ndarray:
        """Get a mask"""
        raise NotImplementedError


class IterDataset(Iterator):
    """Iterable
    """
    def __init__(self, dataset: Dataset):
        self.dataset = dataset
        self.ind = 0
        self.end = len(self.dataset)

    def __next__(self):
        if self.ind == self.end:
            raise StopIteration
        data = self.dataset[self.ind]
        self.ind += 1
        return data



from pathlib import Path
from typing import Dict, List, TypeVar


T = TypeVar('T')

# Keep all ext in lower case
PIL_IMAGE_EXT = [
    '.png',
    '.jpg',
    '.jpeg',
]
TIFFFILE_IMAGE_EXT = [
    '.tif',
    '.tiff',
]
KNOWN_IMAGE_EXT = PIL_IMAGE_EXT + TIFFFILE_IMAGE_EXT
Bundled = List[T]
BundledPath = Bundled[Path]


class DatasetList(list):
    """List of Datasets
    """
    def __init__(self, x):
        for _x in x:
            self.append(_x)

    def set_training(self, val: bool):
        """Iterate config and set all ``training`` to given value

        It only affects those that have ``training`` kwarg.

        """
        attr = 'training'
        for dset in self:
            if hasattr(dset, attr):
                setattr(dset, attr, val)
                # clear cached_property
                if hasattr(dset, 'file_list'):
                    delattr(dset, 'file_list')
                if hasattr(dset, 'anno_dict'):
                    delattr(dset, 'anno_dict')

    def foreach_sample_by_indices(self, indices: Dict[str, List[int]]):
        """Iter Datasets and sample by indices

        Set attribute `_indices`

        Arguments
        ---------
        indices : dictionary
            Key is acronym and value is list of indices
        """
        from .utils import split_dataset_by_indices

        for i, dset in enumerate(self):
            if (name := dset.acronym) in indices:
                _indices = indices[name]
                self[i] = split_dataset_by_indices(dset, _indices)
                setattr(self[i], '_indices', _indices)

# def split_dataset_by_indices(
#     dataset: MaskDatasetProto,
#     indices: Sequence[int],
# ) -> List[MaskDataset]:
#     """Split dataset given indices
#     """
#     return subset(dataset, indices)



"""StarDist data[1]_ is a subset of DSB2018

.. [1] https://github.com/stardist/stardist/releases/download/0.1.0/dsb2018.zip
"""

import os.path
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

import cv2
import numpy as np
import tifffile
from skimage.util import img_as_float32

from ..base import MaskDataset

if TYPE_CHECKING:
    import albumentations


class StarDist(MaskDataset):
    """Dataset for StarDist [1]_, [2]_

    Cell Detection with Star-convex Polygons

    StarDist data is a subset of Data Science Bowl 2018 [3]_

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    training : bool, default: True
        Load training set if True, else load testing one

    Notes
    -----
    - StarDist data is a subset of Data Science Bowl 2018 [3]_. Choose only one,
      do not mix them.
    - ``root_dir`` is not 'dsb2018' even though the archive name is 'dsb2018',
      because it conflicts with the original DSB2018. Make a new directory.
    - All images have grayscale

    References
    ----------
    .. [1] U. Schmidt, M. Weigert, C. Broaddus, and G. Myers, “Cell Detection
       with Star-convex Polygons,” arXiv:1806.03535 [cs], vol. 11071, pp.
       265–273, 2018, doi: 10.1007/978-3-030-00934-2_30.
    .. [2] https://github.com/stardist/stardist/releases/download/0.1.0/dsb2018.zip
    .. [3] https://www.kaggle.com/c/data-science-bowl-2018/

    See Also
    --------
    DSB2018
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = 'StarDist'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        # specific to this dataset
        training: bool = True,
        **kwargs
    ):
        self._root_dir = os.path.join(root_dir, 'dsb2018')
        self._output = output
        self._transforms = transforms
        self._num_samples = num_samples
        # specific to this one here
        self.training = training

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        img = img_as_float32(img)
        return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)

    def get_mask(self, p: Path) -> np.ndarray:
        mask = tifffile.imread(p)
        # originally uint16, which pytorch doesn't like
        return mask.astype(np.int16)

    @cached_property
    def file_list(self) -> List[Path]:
        # Call MaskDataset.root_dir
        parent = 'train' if self.training else 'test'
        return sorted((self.root_dir / parent / 'images').glob('*.tif'))

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        return dict((i, p.parent.with_name('masks') / p.name)
                    for i, p in enumerate(self.file_list))



import os.path
import warnings
from functools import cached_property
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import yaml

from .base import Dataset
from .collections import *
from .types import DatasetList

if TYPE_CHECKING:
    import albumentations


class Config(dict):
    """Construct config from a yaml file

    Parameters
    ----------
    filename : str
        Path to config file
    """
    def __init__(self, filename=None):
        if filename is not None:
            self.filename = filename
            cfg = self._read_yaml()
            for k, v in cfg.items():
                self[k] = v

    def _read_yaml(self) -> dict:
        with open(self.filename, 'r') as f:
            cfg = yaml.safe_load(f)
        return cfg

    @staticmethod
    def from_dict(d: dict):
        return _Config(d)

    def load_datasets(
        self,
        transforms: Optional[Union['albumentations.Compose', Dict[str, 'albumentations.Compose']]] = None,
    ) -> DatasetList:
        """Load multiple datasets from a yaml file

        Note that when you provide a dictionray for ``transforms``, keys should
        be the class names, not their acronyms.

        Parameters
        ----------
        config : configuration object
            Config instance that contains acronyms and arguments to initialize
            each dataset
        transforms : albumentations.Compose or dictionary, optional
            Either apply a single composed transformations for every datasets or
            pass a dictionary that defines transformations for each dataset with
            keys being the class names of collections.

        """
        datasets: List[Dataset] = []
        for dataset, kwargs in self.items():
            if isinstance(transforms, dict):
                if dataset in transforms:
                    exec(f'datasets.append({dataset}(transforms=transforms[dataset], **kwargs))')
                else:
                    exec(f'datasets.append({dataset}(**kwargs))')
            else:
                exec(f'datasets.append({dataset}(transforms=transforms, **kwargs))')
        return DatasetList(datasets)

    @cached_property
    def commonpath(self):
        commonpath = os.path.commonpath([
            p for p in map(lambda x: x['root_dir'], self.values())
        ])
        return commonpath

    def replace_commonpath(self, new: str):
        """Replace common path for all ``root_dir`` with a new one

        All ``root_dir`` should have a commonpath. Do not put trailing '/' at
        the end.

        You made a config with root_dir being relative path. You do not need to
        replace them manually with this method.
        """
        for k in self.keys():
            p = self[k]['root_dir']
            self[k]['root_dir'] = p.replace(self.commonpath, new.rstrip('/'))
        delattr(self, 'commonpath')

    def set_training(self, val: bool):
        """Iterate config and set all ``training`` to given value

        It only affects those that have ``training`` kwarg.

        """
        attr = 'training'
        warnings.warn(f"This method only set those that have `{attr}` kwarg.",
                      stacklevel=2)
        for k in self.keys():
            if attr in self[k]:
                self[k][attr] = val

    def set_ouput(self, val: str):
        """Iterate config and set all ``output`` to given value

        It only affects those that have ``output`` kwarg.

        """
        attr = 'output'
        warnings.warn(f"This method only set those that have `{attr}` kwarg.",
                      stacklevel=2)
        for k in self.keys():
            if attr in self[k]:
                self[k][attr] = val

    def set_grayscale(self, val: bool):
        """Iterate config and set all ``grayscale`` to given value

        It only affects those that have ``grayscale`` kwarg.

        """
        attr = 'grayscale'
        warnings.warn(f"This method only set those that have `{attr}` kwarg.",
                      stacklevel=2)
        for k in self.keys():
            if attr in self[k]:
                self[k][attr] = val


class _Config(Config):
    """Config.from_dict()"""
    def __init__(self, d: dict):
        for k, v in d.items():
            self[k] = v



from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Union

import cv2
import numpy as np
from skimage.util import img_as_float32

from bioimageloader.base import MaskDataset
from bioimageloader.utils import imread_asarray

if TYPE_CHECKING:
    import albumentations


class Cellpose(MaskDataset):
    """Dataset for Cellpose [1]_, [2]_

    Cellpose: a generalist algorithm for cellular segmentation

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'cv2', 'equal', Sequence[float]}, default: 'cv2'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    training : bool, default: True
        Load training set if True, else load testing one.
    gray_is_not_green : bool, default: True
        Proper grayscale. Green channel value will be broadcast to all channels.
    specialized_data : bool, default: False
        Load "specialized data" mentioned in the paper [1]_.

    Notes
    -----
    - Download link is hard to find [3]_
    - It is a complete dataset by itself, meaning that it is not intended to be
      mixed or concatenated with others. It consists of various sources of
      images, not only bioimages but also images of fruits, rocks and etc.
    - All images have 3 channels, but technically they are not RGB. Every images
      have values on the second channel and if there is more signal, then it
      goes to the first one. There is no image that has values on the last
      channel. As a result, when visualized in RGB, they look all green and red.
      In particular, for this reason, grayscale images have signal in the second
      channel and look green. ``gray_is_not_green`` argument address that.
    - Built-in grayscale conversion methods are not correct for this dataset.
      The conversion should be channel-agnostic.
    - Currently, ``gray_is_not_green=False`` and ``grayscale=True`` will reduce
      values of single channel images 1/3 times.

    References
    ----------
    .. [1] C. Stringer, M. Michaelos, and M. Pachitariu, “Cellpose: a generalist
       algorithm for cellular segmentation,” bioRxiv, p. 2020.02.02.931238, Feb.
       2020, doi: 10.1101/2020.02.02.931238.
    .. [2] https://github.com/mouseLand/cellpose
    .. [3] https://www.cellpose.org/dataset


    See Also
    --------
    MaskDataset : Super class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = 'Cellpose'

    names_rg_training = ['000_img.png', '001_img.png', '002_img.png',
                         '003_img.png', '004_img.png', '005_img.png',
                         '006_img.png', '007_img.png', '008_img.png',
                         '009_img.png', '010_img.png', '011_img.png',
                         '012_img.png', '013_img.png', '014_img.png',
                         '015_img.png', '016_img.png', '017_img.png',
                         '018_img.png', '019_img.png', '020_img.png',
                         '021_img.png', '022_img.png', '023_img.png',
                         '024_img.png', '025_img.png', '026_img.png',
                         '027_img.png', '028_img.png', '029_img.png',
                         '030_img.png', '031_img.png', '032_img.png',
                         '033_img.png', '034_img.png', '035_img.png',
                         '036_img.png', '037_img.png', '038_img.png',
                         '039_img.png', '040_img.png', '041_img.png',
                         '042_img.png', '043_img.png', '044_img.png',
                         '045_img.png', '046_img.png', '047_img.png',
                         '048_img.png', '049_img.png', '050_img.png',
                         '051_img.png', '052_img.png', '053_img.png',
                         '054_img.png', '055_img.png', '056_img.png',
                         '057_img.png', '058_img.png', '059_img.png',
                         '060_img.png', '061_img.png', '062_img.png',
                         '063_img.png', '064_img.png', '065_img.png',
                         '066_img.png', '067_img.png', '068_img.png',
                         '069_img.png', '070_img.png', '071_img.png',
                         '072_img.png', '073_img.png', '074_img.png',
                         '075_img.png', '076_img.png', '077_img.png',
                         '078_img.png', '079_img.png', '080_img.png',
                         '081_img.png', '082_img.png', '083_img.png',
                         '084_img.png', '085_img.png', '086_img.png',
                         '087_img.png', '088_img.png', '089_img.png',
                         '090_img.png', '091_img.png', '092_img.png',
                         '093_img.png', '094_img.png', '095_img.png',
                         '096_img.png', '097_img.png', '098_img.png',
                         '099_img.png', '100_img.png', '101_img.png',
                         '102_img.png', '103_img.png', '104_img.png',
                         '105_img.png', '106_img.png', '107_img.png',
                         '108_img.png', '109_img.png', '110_img.png',
                         '111_img.png', '112_img.png', '113_img.png',
                         '114_img.png', '115_img.png', '116_img.png',
                         '117_img.png', '118_img.png', '119_img.png',
                         '120_img.png', '121_img.png', '122_img.png',
                         '144_img.png', '145_img.png', '146_img.png',
                         '147_img.png', '148_img.png', '149_img.png',
                         '150_img.png', '151_img.png', '152_img.png',
                         '153_img.png', '154_img.png', '156_img.png',
                         '157_img.png', '161_img.png', '162_img.png',
                         '167_img.png', '169_img.png', '177_img.png',
                         '178_img.png', '180_img.png', '181_img.png',
                         '182_img.png', '183_img.png', '185_img.png',
                         '186_img.png', '187_img.png', '191_img.png',
                         '192_img.png', '193_img.png', '195_img.png',
                         '197_img.png', '198_img.png', '199_img.png',
                         '200_img.png', '201_img.png', '203_img.png',
                         '205_img.png', '206_img.png', '207_img.png',
                         '209_img.png', '210_img.png', '213_img.png',
                         '215_img.png', '218_img.png', '222_img.png',
                         '223_img.png', '225_img.png', '226_img.png',
                         '227_img.png', '228_img.png', '229_img.png',
                         '230_img.png', '231_img.png', '232_img.png',
                         '233_img.png', '234_img.png', '235_img.png',
                         '236_img.png', '237_img.png', '238_img.png',
                         '239_img.png', '240_img.png', '243_img.png',
                         '244_img.png', '246_img.png', '250_img.png',
                         '261_img.png', '264_img.png', '269_img.png',
                         '270_img.png', '271_img.png', '272_img.png',
                         '273_img.png', '274_img.png', '280_img.png',
                         '283_img.png', '284_img.png', '285_img.png',
                         '286_img.png', '287_img.png', '332_img.png',
                         '337_img.png', '340_img.png', '439_img.png']

    names_rg_test = ['000_img.png', '001_img.png', '002_img.png', '003_img.png',
                     '004_img.png', '005_img.png', '006_img.png', '007_img.png',
                     '008_img.png', '009_img.png', '010_img.png', '011_img.png',
                     '017_img.png', '019_img.png', '021_img.png', '022_img.png']

    idx_sp_train = list(range(89))
    idx_sp_test = list(range(11))

    # num_channels = 2

    def __init__(
        self,
        root_dir: str,
        *,  # only keyword param
        output: str = 'both',
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        grayscale: bool = False,  # optional
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        training: bool = True,
        gray_is_not_green: bool = True,
        specialized_data: bool = False,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_samples = num_samples
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this one here
        self.training = training
        self.gray_is_not_green = gray_is_not_green
        self.specialized_data = specialized_data

        self.names_rg = self.names_rg_training if training else self.names_rg_test
        if specialized_data:
            self.specialized_idx = self.idx_sp_train if training else self.idx_sp_test

    def get_image(self, p: Path) -> np.ndarray:
        img = imread_asarray(p)
        img = img_as_float32(img)
        if p.name in self.names_rg:
            return img
        if self.gray_is_not_green:
            return cv2.cvtColor(img[..., 1], cv2.COLOR_GRAY2RGB)
        return img

    def get_mask(self, p: Path) -> np.ndarray:
        mask = imread_asarray(p)
        return mask

    @cached_property
    def file_list(self) -> List[Path]:
        # Important to decorate with `cached_property` in general
        parent = 'train' if self.training else 'test'
        _file_list = sorted((self.root_dir / parent).glob('*_img.png'))
        if self.specialized_data:
            return [_file_list[i] for i in self.specialized_idx]
        return _file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        # Important to decorate with `cached_property` in general
        parent = 'train' if self.training else 'test'
        _anno_dict = dict((i, v) for i, v in enumerate(
            sorted((self.root_dir / parent).glob('*_masks.png'))))
        if self.specialized_data:
            return dict((i, _anno_dict[i]) for i in self.specialized_idx)
        return _anno_dict



from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Union

import numpy as np
from PIL import Image
from skimage.util import img_as_float32

from ..base import MaskDataset
from ..utils import imread_asarray

if TYPE_CHECKING:
    import albumentations

class BBBC030(MaskDataset):
    """Chinese Hamster Ovary Cells

    The image set consists of 60 Differential Interference Contrast (DIC) images of Chinese Hamster Ovary (CHO) cells.
    The images are taken on an Olympus Cell-R microscope with a 20x lens at the time when the cell initiated their attachment to the bottom of the dish.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    See Also
    --------
    MaskDataset : Super class
    DatasetInterface : Interface

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC030

    """
    # Set acronym
    acronym = 'BBBC030'

    def __init__(
        self,
        root_dir: str,
        *,  # only keyword param
        output: str = 'both',
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        # specific to this dataset
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_samples = num_samples
        # specific to this one here

    def get_image(self, p: Path) -> np.ndarray:
        img = imread_asarray(p)
        return img_as_float32(img)

    def get_mask(self, p: Path) -> np.ndarray:
        mask = imread_asarray(p)
        return mask

    @cached_property
    def file_list(self) -> List[Path]:
        # Important to decorate with `cached_property` in general
        root_dir = self.root_dir
        parent = 'images'
        file_list = sorted(root_dir.glob(f'{parent}/*.png'))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        # Important to decorate with `cached_property` in general
        root_dir = self.root_dir
        parent = 'ground_truth'
        anno_list = sorted(root_dir.glob(f'{parent}/*.png'))
        anno_dict = dict((k, v) for k, v in enumerate(anno_list))
        return anno_dict



from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Union

import cv2
import numpy as np
import tifffile
from skimage.util import img_as_float32

from ..base import MaskDataset

if TYPE_CHECKING:
    import albumentations


class BBBC004(MaskDataset):
    """Synthetic cells

    Biological application

    One of the principal challenges in counting or segmenting nuclei is dealing with clustered nuclei.
    To help assess algorithms' performance in this regard, this synthetic image set consists of five
    subsets with increasing degree of clustering.

    Images

    Five subsets of 20 images each are provided. Each image contains 300 objects, but the objects overlap
    and cluster with different probabilities in the five subsets. The images were generated with the SIMCEP
    simulating platform for fluorescent cell population images (Lehmussola et al., IEEE T. Med. Imaging, 2007 and Lehmussola et al., P. IEEE, 2008).

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC004


    See Also
    --------
    MaskDataset : Super class
    DatasetInterface : Interface
    """
    # Set acronym
    acronym = 'BBBC004'

    def __init__(
        self,
        root_dir: str,
        *,  # only keyword param
        output: str = 'both',
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        # specific to this dataset
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_samples = num_samples
        # specific to this one here

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img_as_float32(img)

    def get_mask(self, p: Path) -> np.ndarray:
        mask = tifffile.imread(p)
        mask = mask[:, :, 2]
        return mask

    @cached_property
    def file_list(self) -> List[Path]:
        # Important to decorate with `cached_property` in general
        root_dir = self.root_dir
        parent = '*_images'
        file_list = sorted(root_dir.glob(f'{parent}/*.tif'), key=lambda x: int(x.stem.split("GRAY")[0]))
        return file_list

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        # Important to decorate with `cached_property` in general
        root_dir = self.root_dir
        parent = '*_foreground'
        anno_list = sorted(root_dir.glob(f'{parent}/*.tif'), key=lambda x: int(x.stem))
        anno_dict = dict((k, v) for k, v in enumerate(anno_list))
        return anno_dict



from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Sequence, Union

import cv2
import numpy as np
from PIL import Image
from skimage.util import img_as_float32

from ..base import Dataset
from ..types import BundledPath
from ..utils import bundle_list, stack_channels_to_rgb

if TYPE_CHECKING:
    import albumentations


class BBBC013(Dataset):
    """Human U2OS cells cytoplasm–nucleus translocation

    The images were acquired at BioImage on the IN Cell Analyzer 3000 using the
    Trafficking Data Analysis Module, with one image per channel (Channel 1 =
    FKHR-GFP; Channel 2 = DNA). Image size is 640 x 640 pixels. Images are
    available in native FRM format or 8-bit BMP.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    grayscale : bool, default: False
        Convert images to grayscale
    grayscale_mode : {'equal', 'cv2', Sequence[float]}, default: 'equal'
        How to convert to grayscale. If set to 'cv2', it follows opencv
        implementation. Else if set to 'equal', it sums up values along channel
        axis, then divides it by the number of expected channels.
    image_ch : {'GFP', 'DNA'}, default: ('GFP', 'DNA')
        Which channel(s) to load as image. Make sure to give it as a Sequence
        when choose a single channel.

    Notes
    -----
    - Two formats are available; FRM and BMP

    References
    ----------
    .. [1] https://bbbc.broadinstitute.org/BBBC013

    See Also
    --------
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Dataset's acronym
    acronym = 'BBBC013'

    def __init__(
        self,
        root_dir: str,
        *,
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        grayscale: bool = False,
        grayscale_mode: Union[str, Sequence[float]] = 'equal',
        # specific to this dataset
        image_ch: Sequence[str] = ('GFP', 'DNA',),
        **kwargs
    ):
        self._root_dir = root_dir
        self._transforms = transforms
        self._num_samples = num_samples
        self._grayscale = grayscale
        self._grayscale_mode = grayscale_mode
        # specific to this dataset
        self.image_ch = image_ch
        if not any([ch in ('GFP', 'DNA') for ch in image_ch]):
            raise ValueError("Set `image_ch` in ('GFP', 'DNA') in sequence")

    def get_image(self, p: Union[Path, BundledPath]) -> np.ndarray:
        if isinstance(p, Path):
            img = np.asarray(Image.open(p))
            img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        else:
            img = stack_channels_to_rgb(Image.open, p)
        return img_as_float32(img)

    @cached_property
    def file_list(self) -> Union[List[Path], List[BundledPath]]:
        root_dir = self.root_dir
        parent = 'BBBC013_v1_images_bmp'
        file_list = sorted(root_dir.glob(f'{parent}/*.BMP'), key=self._sort_key)
        if len(ch := self.image_ch) == 1:
            if ch[0] == 'GFP':
                return file_list[::2]
            elif ch[0] == 'DNA':
                return file_list[1::2]
        return bundle_list(file_list, 2)

    @staticmethod
    def _sort_key(p: Path):
        channel, ind, t, subind = p.stem.split('-')
        return '-'.join([ind, t, subind, channel])



import warnings
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

import cv2
import numpy as np
import tifffile
from pycocotools import coco
from skimage.util import img_as_float32

from ..base import MaskDataset

if TYPE_CHECKING:
    import albumentations


class LIVECell(MaskDataset):
    """LIVECEll: A large-scale dataset for label-free live cell segmentation
    [1]_

    “LIVECell - A large-scale dataset for label-free live cell segmentation” by
    Edlund et. al. 2021 [2]_

    Light microscopy is a cheap, accessible, non-invasive modality that when
    combined with well-established protocols of two-dimensional cell culture
    facilitates high-throughput quantitative imaging to study biological
    phenomena. Accurate segmentation of individual cells enables exploration of
    complex biological questions, but this requires sophisticated imaging
    processing pipelines due to the low contrast and high object density. Deep
    learning-based methods are considered state-of-the-art for most computer
    vision problems but require vast amounts of annotated data, for which there
    is no suitable resource available in the field of label-free cellular
    imaging. To address this gap we present LIVECell, a high-quality, manually
    annotated and expert-validated dataset that is the largest of its kind to
    date, consisting of over 1.6 million cells from a diverse set of cell
    morphologies and culture densities. To further demonstrate its utility, we
    provide convolutional neural network-based models trained and evaluated on
    LIVECell.

    Parameters
    ----------
    root_dir : str
        Path to root directory
    output : {'both', 'image', 'mask'}, default: 'both'
        Change outputs. 'both' returns {'image': image, 'mask': mask}.
    transforms : albumentations.Compose, optional
        An instance of Compose (albumentations pkg) that defines augmentation in
        sequence.
    num_samples : int, optional
        Useful when ``transforms`` is set. Define the total length of the
        dataset. If it is set, it overwrites ``__len__``.
    training : bool, default: True
        Load training set if True, else load testing one
    mask_tif : bool, default: False
        Use saved COCO annotations as tif mask images in a new ./root_dir/masks
        directory. It will greatly improve loading speed. Available after
        calling ``save_coco_to_tif()``.

    Notes
    -----
    - Annotation in MS COCO format [3]_. Parsing it takes time`.
    - Currently not supporting dynamically parsing COCO annotation due to slow
      speed. Pre-parse masks in .tif format by calling ``save_coco_to_tif()``.
    - Validation set is originally separted from training set. Currently they
      are combined ``training=True``.
    - Single cells subsets are not covered

    References
    ----------
    .. [1] https://sartorius-research.github.io/LIVECell/
    .. [2] https://www.nature.com/articles/s41592-021-01249-6
    .. [3] https://cocodataset.org/

    See Also
    --------
    MaskDataset : Super class
    Dataset : Base class
    DatasetInterface : Interface

    """
    # Set acronym
    acronym = 'LIVECell'

    def __init__(
        self,
        root_dir: str,
        *,
        output: str = 'both',
        transforms: Optional['albumentations.Compose'] = None,
        num_samples: Optional[int] = None,
        # specific to this dataset
        training: bool = True,
        mask_tif: bool = False,
        **kwargs
    ):
        self._root_dir = root_dir
        self._output = output
        self._transforms = transforms
        self._num_samples = num_samples
        # specific to this one here
        self.training = training
        self.mask_tif = mask_tif

        if not self.mask_tif:
            msg = ("LIVECell dataset does not currently support dynamically "
                   "parsing annotation in MS COCO format due to performance "
                   "issue. Please create masks in .tif format by calling "
                   "`save_coco_to_tif()` and then set `mask_tif=True`")
            warnings.warn(msg, stacklevel=2)

        if self.mask_tif:
            # check
            mask_train_dir = self.root_dir / 'masks' / 'livecell_train_val_masks'
            mask_test_dir = self.root_dir / 'masks' / 'livecell_test_masks'
            if not mask_train_dir.exists() or not mask_test_dir.exists():
                raise Exception("No masks in .tif format.")
            if not any(mask_train_dir.iterdir()) or not any(mask_test_dir.iterdir()):
                raise Exception("No masks in .tif format.")

    def get_image(self, p: Path) -> np.ndarray:
        img = tifffile.imread(p)
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
        return img_as_float32(img)

    def get_mask(self, p: Path) -> np.ndarray:
        mask = tifffile.imread(p)
        return mask

    @cached_property
    def file_list(self) -> List[Path]:
        # Call MaskDataset.root_dir
        root_dir = self.root_dir
        parent = (root_dir / 'images' / 'livecell_train_val_images' if self.training
                  else root_dir / 'images' / 'livecell_test_images')
        return sorted(parent.glob('*.tif'))

    @cached_property
    def anno_dict(self) -> Dict[int, Path]:
        root_dir = self.root_dir
        # parent = 'masks/livecell_train_val_masks' if self.training else 'masks/livecell_test_masks'
        parent = (root_dir / 'masks' / 'livecell_train_val_masks' if self.training
                  else root_dir / 'masks' / 'livecell_test_masks')
        return dict((k, v) for k, v in enumerate(
            sorted(parent.glob('*.tif'))
        ))

    def save_coco_to_tif(self):
        """Save the masks as tif files

        Read training/val or test annotations from json. Make tif files under
        'masks/livecell_train_val_masks' and 'masks/livecell_test_masks'.

        Initialize a new instance with setting ``mask_tif=True`` to load saved
        masks.
        """
        print("making instances masks and saving as tif files")
        # makedirs
        mask_train_dir = self.root_dir / 'masks' / 'livecell_train_val_masks'
        mask_test_dir = self.root_dir / 'masks' / 'livecell_test_masks'
        mask_train_dir.mkdir(parents=True, exist_ok=True)
        mask_test_dir.mkdir(parents=True, exist_ok=True)
        # training
        self.coco_tr = coco.COCO(self.root_dir  / 'livecell_coco_train.json')
        img_tr = self.coco_tr.loadImgs(self.coco_tr.getImgIds())
        self.coco_val = coco.COCO(self.root_dir / 'livecell_coco_val.json')
        img_val = self.coco_val.loadImgs(self.coco_val.getImgIds())
        self.anno_dictionary = img_val + img_tr
        # loop
        print(f'training total: {len(self.anno_dictionary)}')
        for ind, img in enumerate(self.anno_dictionary, 1):
            try:
                annIds = self.coco_tr.getAnnIds(imgIds=img["id"], iscrowd=None)
                anns = self.coco_tr.loadAnns(annIds)
                mask = self.coco_tr.annToMask(anns[0])
                mask = mask.astype(np.int32)
                for i in range(len(anns)):
                    mask |= self.coco_tr.annToMask(anns[i]) * i
            except:
                annIds = self.coco_val.getAnnIds(imgIds=img["id"], iscrowd=None)
                anns = self.coco_val.loadAnns(annIds)
                mask = self.coco_val.annToMask(anns[0])
                mask = mask.astype(np.int32)
                for i in range(len(anns)):
                    mask |= self.coco_val.annToMask(anns[i]) * i
            tifffile.imsave(mask_train_dir / img['file_name'], mask)
            print(ind, end=' ')
        print("Done!")
        # test
        print("making instances masks and saving as tif files")
        self.coco_te = coco.COCO(self.root_dir / 'livecell_coco_test.json')
        img_te = self.coco_te.loadImgs(self.coco_te.getImgIds())
        self.anno_dictionary = img_te
        print(f'test total: {len(self.anno_dictionary)}')
        for ind, img in enumerate(self.anno_dictionary, 1):
            annIds = self.coco_te.getAnnIds(imgIds=img['id'], iscrowd=None)
            anns = self.coco_te.loadAnns(annIds)
            mask = self.coco_te.annToMask(anns[0])
            mask = mask.astype(np.int32)
            for i in range(len(anns)):
                mask |= self.coco_te.annToMask(anns[i]) * i
            tifffile.imsave(mask_test_dir / img['file_name'], mask)
            print(ind, end=' ')
        print("Done!")
